MODULE transfert_mpi_collectives_mod
  IMPLICIT NONE

  {%- set ranks = ( (1,':'), (2,':,:'), (3,':,:,:'), (4,':,:,:,:') ) %}
  {%- set types = ( ('i','INTEGER','MPI_INTEGER'),
  ('r','REAL','MPI_REAL8'), ('l','LOGICAL','MPI_LOGICAL') ) %}

  INTERFACE bcast_mpi
     MODULE PROCEDURE bcast_mpi_c
     {%- for tn, tp, mpi in types %}
     MODULE PROCEDURE bcast_mpi_{{tn}} {%- for rk, junk in ranks %}, bcast_mpi_{{tn}}{{rk}} {%- endfor %}
     {%- endfor %}
  END INTERFACE bcast_mpi

CONTAINS

  SUBROUTINE gather_field(field_loc,field_glo)
    USE field_mod
    USE domain_mod
    USE mpi_mod
    USE mpipara
    TYPE(t_field),POINTER :: field_loc(:)
    TYPE(t_field),POINTER :: field_glo(:)
    INTEGER, ALLOCATABLE :: mpi_req(:)
    INTEGER, ALLOCATABLE :: status(:,:)
    INTEGER :: ireq,nreq
    INTEGER :: ind_glo,ind_loc

    IF (.NOT. using_mpi) THEN
       DO ind_loc=1,ndomain
          field_glo(ind_loc)%rval4d=field_loc(ind_loc)%rval4d
       ENDDO

    ELSE
       nreq=ndomain
       IF (mpi_rank==0) nreq=nreq+ndomain_glo
       ALLOCATE(mpi_req(nreq))
       ALLOCATE(status(MPI_STATUS_SIZE,nreq))

       ireq=0
       IF (mpi_rank==0) THEN
          DO ind_glo=1,ndomain_glo
             ireq=ireq+1
             CALL MPI_IRECV(field_glo(ind_glo)%rval4d,size(field_glo(ind_glo)%rval4d) , MPI_REAL8 ,   &
                  domglo_rank(ind_glo),domglo_loc_ind(ind_glo), comm_icosa, mpi_req(ireq), ierr)
          ENDDO
       ENDIF

       DO ind_loc=1,ndomain
          ireq=ireq+1
          CALL MPI_ISEND(field_loc(ind_loc)%rval4d,size(field_loc(ind_loc)%rval4d) , MPI_REAL8 ,   &
               0, ind_loc, comm_icosa, mpi_req(ireq), ierr)
       ENDDO

       CALL MPI_WAITALL(nreq,mpi_req,status,ierr)

    ENDIF

  END SUBROUTINE gather_field

  SUBROUTINE bcast_field(field_glo)
    USE field_mod
    USE domain_mod
    USE mpi_mod
    USE mpipara
    TYPE(t_field),POINTER :: field_glo(:)
    INTEGER :: ind_glo

    IF (using_mpi) THEN
       DO ind_glo=1,ndomain_glo
          CALL MPI_BCAST(field_glo(ind_glo)%rval4d, size(field_glo(ind_glo)%rval4d) , MPI_REAL8, 0, comm_icosa, ierr)
       ENDDO
    ENDIF

  END SUBROUTINE bcast_field

  SUBROUTINE scatter_field(field_glo,field_loc)
    USE field_mod
    USE domain_mod
    USE mpi_mod
    USE mpipara
    TYPE(t_field),POINTER :: field_glo(:)
    TYPE(t_field),POINTER :: field_loc(:)
    INTEGER, ALLOCATABLE :: mpi_req(:)
    INTEGER, ALLOCATABLE :: status(:,:)
    INTEGER :: ireq,nreq
    INTEGER :: ind_glo,ind_loc

    IF (.NOT. using_mpi) THEN
       DO ind_loc=1,ndomain
          field_loc(ind_loc)%rval4d=field_glo(ind_loc)%rval4d
       ENDDO

    ELSE
       nreq=ndomain
       IF (mpi_rank==0) nreq=nreq+ndomain_glo
       ALLOCATE(mpi_req(nreq))
       ALLOCATE(status(MPI_STATUS_SIZE,nreq))

       ireq=0
       IF (mpi_rank==0) THEN
          DO ind_glo=1,ndomain_glo
             ireq=ireq+1
             CALL MPI_ISEND(field_glo(ind_glo)%rval4d,size(field_glo(ind_glo)%rval4d) , MPI_REAL8 ,   &
                  domglo_rank(ind_glo),domglo_loc_ind(ind_glo), comm_icosa, mpi_req(ireq), ierr)
          ENDDO
       ENDIF
       
       DO ind_loc=1,ndomain
          ireq=ireq+1
          CALL MPI_IRECV(field_loc(ind_loc)%rval4d,size(field_loc(ind_loc)%rval4d) , MPI_REAL8 ,   &
               0, ind_loc, comm_icosa, mpi_req(ireq), ierr)
       ENDDO
       
       CALL MPI_WAITALL(nreq,mpi_req,status,ierr)

    ENDIF

  END SUBROUTINE scatter_field

  !===================  Broadcast routines for strings ==================!

  SUBROUTINE bcast_mpi_cgen(var,nb)
    USE mpi_mod
    USE mpipara
    CHARACTER(LEN=*),INTENT(INOUT) :: Var
    INTEGER,INTENT(IN) :: nb
    IF (.NOT. using_mpi) RETURN
    CALL MPI_BCAST(Var,nb,MPI_CHARACTER,mpi_master,comm_icosa,ierr)
  END SUBROUTINE bcast_mpi_cgen

  SUBROUTINE bcast_mpi_c(var1)
    CHARACTER(LEN=*),INTENT(INOUT) :: Var1
    CALL bcast_mpi_cgen(Var1,len(Var1))
  END SUBROUTINE bcast_mpi_c
  
  {%- for tn, tp, mpi in types %}

  !============= Broadcast routines for {{tp}} scalars and arrays ============!

  SUBROUTINE bcast_mpi_{{tn}}gen(var,nb)
    USE mpi_mod
    USE mpipara
    INTEGER, INTENT(IN) :: nb
    {{tp}}, DIMENSION(nb), INTENT(INOUT) :: var
    IF (using_mpi) CALL MPI_BCAST(Var, nb, {{mpi}}, mpi_master, comm_icosa, ierr)
  END SUBROUTINE bcast_mpi_{{tn}}gen

  SUBROUTINE bcast_mpi_{{tn}}(var)
    USE mpipara
    {{tp}}, INTENT(INOUT) :: var
    {{tp}}                :: var_tmp(1)
    IF (is_mpi_master) var_tmp(1)=var
    CALL bcast_mpi_{{tn}}gen(var_tmp,1)
    var=var_tmp(1)
  END SUBROUTINE bcast_mpi_{{tn}}

  {%- for rank, shape in ranks %}

  SUBROUTINE bcast_mpi_{{tn}}{{rank}}(var)
    {{tp}}, INTENT(INOUT) :: var({{shape}})
    CALL bcast_mpi_{{tn}}gen(var,size(var))
  END SUBROUTINE bcast_mpi_{{tn}}{{rank}}
  {%- endfor %}
  {%- endfor %}

END MODULE transfert_mpi_collectives_mod
